package seis2viewer;

import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.util.ArrayList;
import net.alomax.seismicity.Seismicity;

public class Seis2viewer
{

	final static public String version = "2.0.04 - 29.06.2013";
	final static public String DEFNAME = "seis2viewer.def";
	final static public boolean Debug = true;
	/*
	 * Main method
	 * Arguments:
	 * - name of the nordic events file to parse
	 */

	public static void main( String[] args )
	{
		System.out.println( "Seis2Viewer " + version );
		if( args.length == 0 )
		{
			printOptions();
			System.exit( -1 );
		}

		/* Define options */
		boolean includeAll = false;
		boolean forceGlobalMode = false;
		File nfile = null;
		boolean includeMap = false;

		/* Define default seismicityviewer options */
		String[] defSeismicityDefaults = new String[]
		{
			"-view.invertcolors=YES",
			"-init.rotation=0,0,0",
			"-show.axes3d=NO",
			"-color.grid=black",
			"-event.depthcolormap=0,blue",
			"-show.magnitude=YES"
		};


		/* Check input options */
		for( int i = 0; i < args.length; i++ )
		{
			/* Check argument is an option or the file name */
			if( args[i].charAt( 0 ) == '-' )
			{
				/* This is an option. Check if known and valid */
				if( args[i].substring( 1 ).trim().toLowerCase().equalsIgnoreCase( "ia" ) )
				{
					includeAll = true;
				}
				else if( args[i].substring( 1 ).trim().toLowerCase().equalsIgnoreCase( "gl" ) )
				{
					forceGlobalMode = true;
				}
				else
				{
					/* Unknown option */
					System.err.println( "Invalid option: " + args[i] );
					printOptions();
					System.exit( -1 );
				}
			}
			else
			{
				/* This should be a file. Check if it is the last option */
				if( i != ( args.length - 1 ) )
				{
					/* This is not the last option so it should not be a file */
					System.err.println( "Invalid option: " + args[i] );
					printOptions();
					System.exit( -1 );
				}
				nfile = new File( args[i] );
			}
		}
		if( nfile == null )
		{
			System.err.println( "No input file." );
			printOptions();
			System.exit( -1 );
		}

		/* Look for input file */
		if( !nfile.exists() )
		{
			System.err.println( "Unable to find input file: " + args[0] );
			System.exit( -1 );
		}
		else if( !nfile.canRead() )
		{
			System.err.println( "Unable to read input file: " + args[0] );
			System.exit( -1 );
		}

		/* Read entire file to a string */
		String rawnordic = null;
		try
		{
			/* Create buffer */
			char[] buffer = null;
			try
			{
				buffer = new char[(int) nfile.length()];
			}
			catch( Exception e )
			{
				System.err.println( "File too large: " + args[0] );
				System.exit( -1 );
			}
			/* Open stream */
			FileReader Finput = new FileReader( nfile );
			/* Read data */
			Finput.read( buffer );
			/* Close stream */
			Finput.close();
			/* Convert to string */
			rawnordic = new String( buffer );
		}
		catch( Exception e )
		{
			System.err.println( "Unable to read input file: " + args[0] );
			e.printStackTrace();
			System.exit( -1 );
		}

		/* Create eventgroup */
		EventGroup events = null;
		try
		{
			events = new EventGroup( rawnordic, includeAll );
			System.out.println( "Found " + events.size() + " events." );
		}
		catch( Exception e )
		{
			System.err.println( "Critical error parsing nordic file:" + e.getMessage() );
			System.exit( -1 );
		}

		/* Create output file */
		String[] infname = nfile.getName().split( "[.]+" );
		String tempoutfname = "";
		for( int i = 0; i < infname.length - 1; i++ )
			tempoutfname += infname[i];
		final String outfname = tempoutfname;
		System.out.println( "Output file: " + outfname + ".hyp" );
		try
		{
			FileWriter Foutput = new FileWriter( outfname + ".hyp" );
			Foutput.write( events.toNLL() );
			Foutput.flush();
			Foutput.close();
		}
		catch( Exception e )
		{
			System.err.println( "Error creating output events file: " + outfname + ".hyp" );
			e.printStackTrace();
			System.exit( -1 );
		}


		/* Try to find configuration file */
		/* in the local folder */
		File dfile = new File( DEFNAME );
		if( !dfile.exists() || !dfile.canRead() )
		{
			/* Not in local folder, maybe in DAT folder */
			String seisantop = System.getenv( "SEISAN_TOP" );
			if( seisantop == null )
			{
				System.err.println( "Could not find environmental variable: SEISAN_TOP\n"
						+ "Continuing with default parameters." );
				dfile = null;
			}
			else
			{
				dfile = new File( seisantop + "/DAT/" + DEFNAME );
				if( !dfile.exists() || !dfile.canRead() )
				{
					System.err.println( "Could not find definitions file: " + DEFNAME
							+ "Continuing with default parameters." );
					dfile = null;
				}
			}
		}

		/* Create options list */
		String[] options = defSeismicityDefaults;
		if( dfile != null )
		{
			/* Read .def file */
			String rawdefinitions = null;
			try
			{
				char buffer[] = new char[(int) dfile.length()];
				FileReader Finput = new FileReader( dfile );
				Finput.read( buffer );
				Finput.close();
				rawdefinitions = new String( buffer );
			}
			catch( Exception e )
			{
				System.err.println( "Error reading definitions file." );
				e.printStackTrace();
			}
			/* Parse raw definitions */
			if( rawdefinitions != null )
			{
				ArrayList<String> tempdefs = new ArrayList<String>();
				String[] definitions = rawdefinitions.split( "[\\n]+" );
				if( definitions.length > 0 )
				{
					for( int i = 0; i < definitions.length; i++ )
					{
						String curdef = definitions[i].trim().replaceAll( "\\s", "" );
						if( curdef.isEmpty() || curdef.charAt( 0 ) == '#' )
							continue;
						int posdot = curdef.indexOf( "seismicity." );
						if( posdot < 0 )
							continue;
						posdot += ( "seismicity." ).length();
						String parsedDef = curdef.substring( posdot );
						tempdefs.add( "-" + parsedDef );
						/* Check if options is a map file */
						if( parsedDef.contains( "lines." ) )
							includeMap = true;
					}
				}
				if( !tempdefs.isEmpty() )
				{
					options = new String[tempdefs.size()];
					for( int i = 0, m = tempdefs.size(); i < m; i++ )
						options[i] = tempdefs.get( i );
				}
			}
		}

		/* Create grid file */
		System.out.println( "Grid header file: " + outfname + ".hdr" );
		try
		{
			FileWriter Foutput = new FileWriter( outfname + ".hdr" );
			if( !forceGlobalMode )
				Foutput.write( events.toNLLGrid( includeMap ) );
			else
				Foutput.write( events.toNLLGrid( false ) );
			Foutput.flush();
			Foutput.close();
		}
		catch( Exception e )
		{
			System.err.println( "Error creating output grid header file" + outfname + ".hdr" );
			e.printStackTrace();
			//System.exit( -1 );
		}


		/* Start seismicityviewer */
		String[] parameters = new String[1 + options.length];
		parameters[0] = outfname + ".hyp";
		System.arraycopy( options, 0, parameters, 1, options.length );
		if( Debug )
		{
			System.out.println( "SeismicityViewer options:" );
			for( int i = 0; i < parameters.length; i++ )
				System.out.println( parameters[i] );
		}
		Seismicity seismicity = new Seismicity( parameters );

	}

	private static void printOptions()
	{
		System.out.println( "Usage: java -jar seis2viewer.jar <options> <nordic file>\n"
				+ "\n"
				+ "Options:\n"
				+ "\t -ia\t\t: Include all hypocenters. Default is to include only main solutions.\n"
				+ "\t -gl\t\t: Force global mode." );
	}
}
